# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from binascii import unhexlify
from datetime import datetime, timedelta
from struct import pack
import unittest

from apns_worker.apns import Feedback


_token1 = '1ba97ad1311307c189696e2369c89fa83d652611a6e3c7370881289e45668fd3'
_token2 = '4c23f042050f48da350cb1079d61189cf7a47cb8df087429c0b2da65226cbecc'


class FeedbackTestCase(unittest.TestCase):
    _when = datetime(2015, 9, 1)
    _timestamp = int((_when - datetime(1970, 1, 1)).total_seconds())

    def test_parse_empty(self):
        buf = b''
        feedback, remainder = Feedback.parse(buf)

        self.assertEqual(feedback, None)
        self.assertEqual(remainder, buf)

    def test_parse_partial_1(self):
        buf = pack('!I', self._timestamp)
        feedback, remainder = Feedback.parse(buf)

        self.assertEqual(feedback, None)
        self.assertEqual(remainder, buf)

    def test_parse_partial_2(self):
        buf = pack('!IH', self._timestamp, 32)
        feedback, remainder = Feedback.parse(buf)

        self.assertEqual(feedback, None)
        self.assertEqual(remainder, buf)

    def test_parse_partial_3(self):
        buf = pack('!IH30s', self._timestamp, 32, unhexlify(_token1))
        feedback, remainder = Feedback.parse(buf)

        self.assertEqual(feedback, None)
        self.assertEqual(remainder, buf)

    def test_parse_one(self):
        buf = pack('!IH32s', self._timestamp, 32, unhexlify(_token1))
        feedback, remainder = Feedback.parse(buf)

        self.assertEqual(feedback.token, _token1)
        self.assertEqual(feedback.when, self._when)
        self.assertEqual(remainder, b'')

    def test_parse_leftovers(self):
        buf = pack('!IH32sI', self._timestamp, 32, unhexlify(_token1), 0x01020304)
        feedback, remainder = Feedback.parse(buf)

        self.assertEqual(feedback.token, _token1)
        self.assertEqual(feedback.when, self._when)
        self.assertEqual(remainder, b'\x01\x02\x03\x04')

    def test_parse_two(self):
        buf = pack(
            '!IH32sIH32s',
            self._timestamp, 32, unhexlify(_token1),
            self._timestamp + 1, 32, unhexlify(_token2)
        )
        f1, remainder = Feedback.parse(buf)
        f2, remainder = Feedback.parse(remainder)

        self.assertEqual(f1.token, _token1)
        self.assertEqual(f1.when, self._when)
        self.assertEqual(f2.token, _token2)
        self.assertEqual(f2.when, self._when + timedelta(seconds=1))
        self.assertEqual(remainder, b'')



from __future__ import unicode_literals, absolute_import

from binascii import hexlify, unhexlify
from calendar import timegm
from collections import namedtuple
from datetime import datetime, timedelta
from importlib import import_module
from itertools import repeat
import json
import logging
from struct import unpack

from six import python_2_unicode_compatible
from six.moves import map, range

from .data import Notification
from .queue import NotificationQueue


logger = logging.getLogger(__name__)


class ApnsManager(object):
    """
    Top-level object for sending Apple push notifications.

    One instance of this object manages a single connection to the Apple push
    notification service and a single queue of notifications to send. For most
    purposes, a single global instance should be sufficient. For high volumes,
    it may be worthwhile to create multiple instances and distribute messages
    across them.

    :param str key_path: Path to your PEM-encoded APNs client key.
    :param str cert_path: Path to your PEM-encoded APNs client certificate.

    :param str environment: APNs environment: `'sandbox'` or `'production'`.

    :param str backend_path: Module path to a subclass of
        :class:`apns_worker.backend.base.Backend`. The backend provides the
        network access and concurrency.

    :param int message_grace: Number of seconds to hold on to a delivered
        message before assuming that it was successful.

    :param error_handler: An optional function to process delivery errors. The
        function should take one argument, which will be an
        :class:`~apns_worker.Error`.

    """
    def __init__(self, key_path, cert_path,
                 environment='production',
                 backend_path='apns_worker.backend.threaded.Backend',
                 message_grace=5, error_handler=None):

        self._queue = NotificationQueue(grace=message_grace)
        self._backend = self._load_backend(backend_path, environment, key_path, cert_path, error_handler)

        self._backend.start()

    def _load_backend(self, path, environment, key_path, cert_path, error_handler):
        path, name = path.rsplit('.', 1)
        mod = import_module(path)
        backend_cls = getattr(mod, name)

        return backend_cls(self._queue, environment, key_path, cert_path, error_handler)

    #
    # Client APIs
    #

    def send_aps(self, tokens, alert=None, badge=None, sound=None, content_available=None, category=None):
        """
        A convenience API to send a standard notification.

        If you're sending a simple notification with a standard payload, you
        can use this shortcut API instead of constructing a :class:`Message`
        manually. Specify any arguments that you wish to include in the `'aps'
        dictionary
        <https://developer.apple.com/library/ios/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/Chapters/ApplePushService.html#//apple_ref/doc/uid/TP40008194-CH100-SW1>`_
        of the payload.

        :param tokens:
        :type tokens: list of str
        :param alert:
        :type alert: str or dictionary
        :param int badge:
        :param str sound:
        :param bool content_available:
        :param str category:

        """
        aps = {}
        if alert is not None:
            aps['alert'] = alert
        if badge is not None:
            aps['badge'] = badge
        if sound is not None:
            aps['sound'] = sound
        if content_available:
            aps['content-available'] = 1
        if category is not None:
            aps['category'] = category

        message = Message(tokens, {'aps': aps})

        return self.send_message(message)

    def send_message(self, message):
        """
        Queues a message for delivery.

        :param message:
        :type message: :class:`~apns_worker.Message`

        """
        self._queue.append(message)

    def flush_messages(self):
        """
        Wait until all queued messages have been delivered.

        This will not return until all messages have been presumed successful
        (according to the delivery grace period).

        The only reason to call this is to make sure the queue is empty before
        allowing a process to terminate.

        """
        delay = self._queue.purge_expired()
        while not self._queue.is_empty():
            self._backend.sleep(delay)
            delay = self._queue.purge_expired()

    def get_feedback(self, callback):
        """
        Start retrieving tokens from the APNs feedback service.

        This will deliver :class:`~apns_worker.Feedback` items to the callback
        asynchronously.

        :param callback: A function that takes a single
            :class:`~apns_worker.Feedback` object. The callback will be called
            zero or more times.

        """
        self._backend.start_feedback(callback)


class Message(object):
    """
    A single push notification to be sent to one or more devices.

    :param list tokens: A list of hex-encoded device tokens.
    :param dict payload: `Payload dictionary
        <https://developer.apple.com/library/ios/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/Chapters/ApplePushService.html#//apple_ref/doc/uid/TP40008194-CH100-SW1>`_.
    :param datetime expiration: An expiration time (optional). If this is a
        naive datetime, it is assumed to be UTC.
    :param int priority: Notification priority (optional). According to the
        current docs, 10 = send now and 5 = send when convenient.

    This validates arguments fairly aggressively and may raise standard
    exceptions.

    """
    def __init__(self, tokens, payload, expiration=None, priority=None):
        self._tokens = tokens
        self._payload = payload
        self._expiration = expiration
        self._priority = priority

        self._validate()

    def _validate(self):
        self._validate_tokens()
        self._validate_payload()
        self._validate_expiration()
        self._validate_priority()

    def _validate_tokens(self):
        self._encoded_tokens = list(map(unhexlify, self._tokens))

    def _validate_payload(self):
        self._encoded_payload = json.dumps(self.payload, ensure_ascii=False, separators=(',', ':')).encode('utf-8')

    def _validate_expiration(self):
        if self.expiration is not None:
            self._encoded_expiration = int(timegm(self.expiration.utctimetuple()))
        else:
            self._encoded_expiration = None

    def _validate_priority(self):
        if self.priority is None:
            pass
        elif isinstance(self.priority, int) and self.priority in range(0, 256):
            pass
        else:
            raise TypeError("Priority must be an integer in [0, 255] or None")

    @property
    def tokens(self):
        return self._tokens

    @property
    def payload(self):
        return self._payload

    @property
    def expiration(self):
        return self._expiration

    @property
    def priority(self):
        return self._priority

    def notifications(self, idents=None):
        """
        Generates a sequence of serializable notifications.

        This generates a sequence of Notification objects, each of which can be
        serialized onto the wire.

        :param iterable idents: An optional iterable of 32-bit identifiers for
            the notifications. :func:`itertools.count` is useful here.

        """
        if idents is None:
            idents = repeat(None)

        for encoded_token in self._encoded_tokens:
            yield Notification(self, encoded_token, next(idents))


@python_2_unicode_compatible
class Error(namedtuple('Error', ['status', 'message', 'token'])):
    """
    Represents a delivery error returned by APNs.

    These are only generated for unrecoverable errors that the client might
    want to know about.

    .. attribute:: status

        The `APNs status code
        <https://developer.apple.com/library/ios/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/Chapters/CommunicatingWIthAPS.html#//apple_ref/doc/uid/TP40008194-CH101-SW4>`_.

    .. attribute:: description

        A human-readable description of the error.

    .. attribute:: message

        The :class:`~apns_worker.Message` that generated the error.

    .. attribute:: token

        The specific device token that generated the error.

    """
    ERR_PROCESSING = 1
    ERR_NO_TOKEN = 2
    ERR_NO_TOPIC = 3
    ERR_NO_PAYLOAD = 4
    ERR_TOKEN_SIZE = 5
    ERR_TOPIC_SIZE = 6
    ERR_PAYLOAD_SIZE = 7
    ERR_TOKEN_INVAL = 8
    ERR_UNKNOWN = 255

    descriptions = {
        ERR_PROCESSING: "Processing error",
        ERR_NO_TOKEN: "Missing device token",
        ERR_NO_TOPIC: "Missing topic",
        ERR_NO_PAYLOAD: "Missing payload",
        ERR_TOKEN_SIZE: "Invalid token size",
        ERR_TOPIC_SIZE: "Invalid topic size",
        ERR_PAYLOAD_SIZE: "Invalid payload size",
        ERR_TOKEN_INVAL: "Invalid token",
    }

    def __str__(self):
        return "APNs error {0}: {1}".format(self.status, self.description)

    @property
    def description(self):
        return self.descriptions.get(self.status, "Unknown")


class Feedback(namedtuple('Feedback', ['token', 'when'])):
    """
    A single record from the APNs feedback service.

    .. attribute:: token

        A hex-encoded device token that can not receive notifications.

    .. attribute:: when

        The time at which this device stopped receiving notifications as a
        naive UTC datetime. If the device was registered with your service
        after this time, you can ignore this feedback.

    """
    _epoch = datetime(1970, 1, 1)

    @classmethod
    def parse(cls, buf):
        """
        Parses one Feedback object from the head of a byte string.

        :returns: A two-tuple with the parsed :class:`~apns_worker.Feedback`
            and the remaining bytes. The first element of the tuple will be
            `None` if it could not be parsed.
        :rtype: (:class:`~apns_worker.Feedback`, bytes)

        """
        feedback = None
        remainder = buf

        if len(buf) >= 6:
            timestamp, token_len = unpack('!IH', buf[:6])
            total_len = 6 + token_len

            if len(buf) >= total_len:
                token = hexlify(buf[6:total_len]).decode('ascii')
                when = cls._epoch + timedelta(seconds=timestamp)

                feedback = cls(token, when)
                remainder = buf[total_len:]

        return (feedback, remainder)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from datetime import datetime, timedelta
from threading import Condition
import unittest

from apns_worker.apns import Message
from apns_worker.backend.base import Backend
from apns_worker.datetime import Now
from apns_worker.queue import NotificationQueue


_token1 = '1111111111111111111111111111111111111111111111111111111111111111'
_token2 = '2222222222222222222222222222222222222222222222222222222222222222'
_token3 = '3333333333333333333333333333333333333333333333333333333333333333'


class QueueTestCase(unittest.TestCase):
    def setUp(self):
        super(QueueTestCase, self).setUp()

        self.queue = NotificationQueue(grace=10)
        self.backend = TestBackend(self.queue)

    def tearDown(self):
        del self.queue
        del self.backend

        super(QueueTestCase, self).tearDown()

    def test_append(self):
        message = Message([_token1, _token2], {})

        self.queue.append(message)

        self.assertEqual(len(self.queue._queue), 2)
        self.assertTrue(all(item.expires is None for item in self.queue._queue))
        self.assertEqual(self.backend.notifies, 1)

    def test_claim_empty(self):
        notif = self.queue.claim()

        self.assertTrue(notif is None)

    def test_claim(self):
        message = Message([_token1, _token2], {})

        self.queue.append(message)
        notif = self.queue.claim()

        self.assertTrue(notif is not None)
        self.assertEqual(len(self.queue._queue), 2)
        self.assertEqual(self.queue._next, 1)
        self.assertTrue(self.queue._queue[0].expires is not None)
        self.assertTrue(self.queue._queue[1].expires is None)

    def test_claim_all(self):
        message = Message([_token1, _token2], {})

        self.queue.append(message)
        self.queue.claim()
        self.queue.claim()
        notif = self.queue.claim()

        self.assertTrue(notif is None)
        self.assertEqual(len(self.queue._queue), 2)
        self.assertEqual(self.queue._next, 2)
        self.assertTrue(all(item.expires is not None for item in self.queue._queue))

    def test_unclaim_empty(self):
        message = Message([_token1, _token2], {})
        ok = self.queue.unclaim(next(message.notifications()))

        self.assertFalse(ok)
        self.assertEqual(len(self.queue._queue), 0)
        self.assertEqual(self.queue._next, 0)

    def test_unclaim_last(self):
        message = Message([_token1, _token2], {})

        self.queue.append(message)
        self.queue.claim()
        notif = self.queue.claim()
        ok = self.queue.unclaim(notif)

        self.assertTrue(ok)
        self.assertEqual(len(self.queue._queue), 2)
        self.assertEqual(self.queue._next, 1)

    def test_unclaim_invalid(self):
        message = Message([_token1, _token2], {})

        self.queue.append(message)
        notif = self.queue.claim()
        self.queue.claim()
        ok = self.queue.unclaim(notif)

        self.assertFalse(ok)
        self.assertEqual(len(self.queue._queue), 2)
        self.assertEqual(self.queue._next, 2)

    def test_backtrack_empty(self):
        self.queue.backtrack(0)

        self.assertEqual(len(self.queue._queue), 0)
        self.assertEqual(self.queue._next, 0)

    def test_backtrack_all(self):
        message = Message([_token1, _token2, _token3], {})

        self.queue.append(message)
        self.queue.claim()
        self.queue.claim()
        self.queue.backtrack(0)

        self.assertEqual(len(self.queue._queue), 2)
        self.assertEqual(self.queue._next, 0)
        self.assertTrue(all(item.expires is None for item in self.queue._queue))
        self.assertEqual(self.backend.notifies, 2)

    def test_purge_none(self):
        message = Message([_token1, _token2, _token3], {})

        self.queue.append(message)
        self.queue.claim()
        self.queue.claim()
        delay = self.queue.purge_expired()

        self.assertEqual(len(self.queue._queue), 3)
        self.assertEqual(self.queue._next, 2)
        self.assertLessEqual(delay, self.queue._grace)

    def test_purge(self):
        start = datetime(2015, 1, 1)
        message = Message([_token1, _token2, _token3], {})

        self.queue.append(message)
        with Now(start):
            self.queue.claim()
        with Now(start + timedelta(seconds=5)):
            self.queue.claim()

        with Now(start + timedelta(seconds=self.queue._grace + 1)):
            delay = self.queue.purge_expired()

        self.assertEqual(len(self.queue._queue), 2)
        self.assertEqual(self.queue._next, 1)
        self.assertLessEqual(delay, self.queue._grace)

    def test_purge_all(self):
        start = datetime(2015, 1, 1)
        message = Message([_token1, _token2, _token3], {})

        self.queue.append(message)
        with Now(start + timedelta(seconds=5)):
            self.queue.claim()
            self.queue.claim()
            self.queue.claim()

        with Now(start + timedelta(seconds=self.queue._grace + 6)):
            delay = self.queue.purge_expired()

        self.assertEqual(len(self.queue._queue), 0)
        self.assertEqual(self.queue._next, 0)
        self.assertEqual(delay, self.queue._grace)

    def test_auto_purge(self):
        start = datetime(2015, 1, 1)
        self.queue._auto_purge_at = start + timedelta(seconds=self.queue._grace)

        with Now(start):
            self.queue.append(Message([_token1], {}))
            self.queue.claim()

        with Now(start + timedelta(seconds=self.queue._grace + 1)):
            self.queue.append(Message([_token2], {}))

        self.assertEqual(len(self.queue._queue), 1)
        self.assertEqual(self.queue._next, 0)

    def test_is_empty_new(self):
        self.assertTrue(self.queue.is_empty())

    def test_not_empty(self):
        message = Message([_token1, _token2, _token3], {})

        self.queue.append(message)

        self.assertFalse(self.queue.is_empty())


class TestBackend(Backend):
    def __init__(self, queue):
        self.lock = Condition()
        queue._set_backend(self)

        # Statistics
        self.notifies = 0

    def start(self):
        pass

    def stop(self):
        pass

    def start_feedback(self):
        pass

    def queue_lock(self):
        return self.lock

    def queue_notify(self):
        self.notifies += 1

        self.lock.notify_all()

    def sleep(self):
        pass



from __future__ import unicode_literals, absolute_import

from collections import deque
from datetime import timedelta
from threading import RLock

from six.moves import map, range

from .datetime import now


class NotificationQueue(object):
    """
    A special queue for Notification objects.

    When a notification is claimed from the queue, we retain a reference to it
    with an expiration. Once expired, sent notifications are eventually purged.
    However, if a delivery failure is subsequently detected, the
    :meth:`~apns_worker.queue.NotificationQueue.backtrack` method can be used
    to rewind the queue to the first notification that failed.

    :param int grace: Seconds to leave a claimed notification in the queue
        before purging it.

    """
    def __init__(self, grace):
        self._grace = grace

        self._queue = deque()
        self._next = 0
        self._idents = _gen_identifiers()
        self._backend = self.DummyBackend()

        self._auto_purge_at = now() + timedelta(seconds=grace)

    def append(self, message):
        """
        Queues a message for delivery.

        :param message: A single message to queue for delivery.
        :type message: :class:`~apns_worker.Message`

        """
        with self._backend.queue_lock():
            self._queue.extend(map(QueuedNotification, message.notifications(self._idents)))
            self._backend.queue_notify()

        self._auto_purge()

    def claim(self):
        """
        Returns the next notification to be sent.

        The returned notification is provisionally removed from the queue, but
        can be restored with a timely call to
        :meth:`~apns_worker.queue.NotificationQueue.backtrack` or
        :meth:`~apns_worker.queue.NotificationQueue.unclaim`.

        :rtype: :class:`~apns_worker.data.Notification`.

        """
        notification = None

        with self._backend.queue_lock():
            if self._next < len(self._queue):
                queued = self._queue[self._next]
                queued.expires = now() + timedelta(seconds=self._grace)
                notification = queued.notification
                self._next += 1

        return notification

    def unclaim(self, notification):
        """
        Restores the most recently claimed notification to the queue.

        :type notification: :class:`~apns_worker.data.Notification`

        :returns: `True` if the notification could be unclaimed, `False`
            otherwise.
        :rtype: bool

        """
        with self._backend.queue_lock():
            success = False

            if self._next > 0:
                qn = self._queue[self._next - 1]
                if qn.notification == notification:
                    qn.expires = None
                    self._next -= 1
                    success = True

            return success

    def backtrack(self, ident):
        """
        Returns claimed notifications to the queue.

        All notifications after `ident` are known to have failed and need to be
        re-queued. All notifications before `ident` are now known to have
        succeeded.

        :param int ident: Ident of the first failed (or last successful)
            notification.

        :returns: The notification with the given ident, if found.
        :rtype: :class:`~apns_worker.data.Notification` or None.

        """
        notification = None

        with self._backend.queue_lock():
            queue = self._queue
            i = (self._next - 1) if (self._next > 0) else 0

            # Try to find the failed notification. We'll stop if we hit the
            # beginning of the queue.
            while (i > 0) and (queue[i].notification.ident != ident):
                i -= 1

            if i < len(queue) and (queue[i].notification.ident == ident):
                notification = queue[i].notification
                i += 1

            # Everything else either succeeded or failed permanently.
            for j in range(i):
                queue.popleft()

            # Reset all expirations.
            for item in queue:
                item.expires = None

            self._next = 0

            self._backend.queue_notify()

        return notification

    def purge_expired(self):
        """
        Permanently removes notifications that have expired.

        This is called auomatically as oppotunities arise, but can also be
        called manually, for instance to drain the queue for termination.

        :returns: A recommended number of seconds to wait until the next call.
        :rtype: float

        """
        with self._backend.queue_lock():
            _now = now()
            is_expired = lambda qn: qn.is_claimed() and (qn.expires <= _now)

            while (len(self._queue) > 0) and is_expired(self._queue[0]):
                self._queue.popleft()
                self._next -= 1

            if len(self._queue) > 0 and self._queue[0].is_claimed():
                delay = (self._queue[0].expires - _now).total_seconds()
            else:
                delay = self._grace

        return max(delay, 1.0)

    def has_unclaimed(self):
        """
        Returns `True` if the queue has any unclaimed items.

        :rtype: bool

        """
        with self._backend.queue_lock():
            has_unclaimed = (self._next < len(self._queue))

        return has_unclaimed

    def is_empty(self):
        """
        Returns `True` if the queue has no items.

        This includes claimed and unclaimed items. Use this with
        :meth:`~apns_worker.queue.NotificationQueue.purge_expired` to wait for
        a decommissioned queue to drain.

        :rtype: bool

        """
        with self._backend.queue_lock():
            is_empty = (len(self._queue) == 0)

        return is_empty

    #
    # Internal
    #

    class DummyBackend(object):
        def __init__(self):
            self.lock = RLock()

        def queue_lock(self):
            return self.lock

        def queue_notify(self):
            pass

    def _set_backend(self, backend):
        self._backend = backend

    def _auto_purge(self):
        _now = now()

        with self._backend.queue_lock():
            if _now > self._auto_purge_at:
                delay = self.purge_expired()
                self._auto_purge_at = _now + timedelta(seconds=delay)


def _gen_identifiers():
    """ Generates sequential 32-bit notification identifiers. """
    while True:
        for i in range(2 ** 32):
            yield i


class QueuedNotification(object):
    __slots__ = ['notification', 'expires']

    def __init__(self, notification):
        self.notification = notification
        self.expires = None

    def is_claimed(self):
        return (self.expires is not None)



"""
APNs data structures with serialization.
"""
from __future__ import unicode_literals, absolute_import

from binascii import hexlify
from struct import pack

from six import python_2_unicode_compatible


@python_2_unicode_compatible
class Notification(object):
    """
    An internal representation of a single notification to send.

    :param message: The original message.
    :type message: :class:`apns_worker.Message`

    :param bytes encoded_token: Binary representation of a device token.
    :param int ident: 32-bit notification identifier.

    """
    def __init__(self, message, encoded_token, ident):
        self.message = message
        self.encoded_token = encoded_token
        self.ident = ident

    def __str__(self):
        return "{0} -> {1}".format(self.message._encoded_payload, self.token)

    @property
    def token(self):
        """
        This notification's hex-encoded token.

        :rtype: str

        """
        return hexlify(self.encoded_token)

    def frame(self):
        """
        Renders this notification to an APNs frame.

        :returns: A complete frame, ready to be put on the wire.
        :rtype: bytes

        """
        encoded_payload = self.message._encoded_payload
        encoded_expiration = self.message._encoded_expiration
        priority = self.message.priority

        content = b''.join([
            pack('!BH32s', 1, 32, self.encoded_token),
            self._pack_data(2, encoded_payload),
            pack('!BHI', 3, 4, self.ident) if (self.ident is not None) else b'',
            pack('!BHI', 4, 4, encoded_expiration) if (encoded_expiration is not None) else b'',
            pack('!BHB', 5, 1, priority) if (priority is not None) else b'',
        ])

        frame = pack('!BI', 2, len(content)) + content

        return frame

    def _pack_data(self, item_id, data):
        """ Packs variable size data. """
        packed = b''

        if data is not None:
            length = len(data)
            packed = pack('!BH{0}s'.format(length), item_id, length, data)

        return packed



from __future__ import unicode_literals

from binascii import hexlify, unhexlify
from collections import namedtuple
from datetime import datetime, timedelta
from itertools import takewhile, repeat
import json
import logging
import socket
import struct
from threading import Condition
from time import sleep
import unittest

try:
    from unittest import mock
except ImportError:
    import mock

import six

from apns_worker import ApnsManager, Message
from apns_worker.backend.threaded import Connection


_token1 = '1111111111111111111111111111111111111111111111111111111111111111'
_token2 = '2222222222222222222222222222222222222222222222222222222222222222'
_token3 = '3333333333333333333333333333333333333333333333333333333333333333'


class BackendTestCase(unittest.TestCase):
    _when = datetime(2015, 9, 1)
    _timestamp = int((_when - datetime(1970, 1, 1)).total_seconds())

    @classmethod
    def setUpClass(cls):
        super(BackendTestCase, cls).setUpClass()

        logger = logging.getLogger('apns_worker')
        logger.setLevel(logging.WARNING)
        logger.addHandler(logging.StreamHandler())

    def setUp(self):
        super(BackendTestCase, self).setUp()

        self._connection_patch = mock.patch('apns_worker.backend.threaded._new_connection', self.new_connection)
        self._connection_patch.start()

        self._apns = None
        self.connection_inbuf = None
        self.connections = []
        self.apns_error = None
        self.feedbacks = []

    def tearDown(self):
        if self._apns is not None:
            self._apns._backend.stop()
            self._apns = None

        self._connection_patch.stop()
        del self._connection_patch

        super(BackendTestCase, self).tearDown()

    def test_wait_for_notification(self):
        self.assertFalse(self.apns._backend.thread.connection.is_opened)

    def test_send_aps(self):
        kwargs = {
            'alert': 'Alert!',
            'badge': 5,
            'sound': 'Meep Meep!',
            'content_available': True,
            'category': 'stuff',
        }
        self.apns.send_aps([_token1], **kwargs)

        sleep(0.1)

        frame = self.sent_frames[0]
        payload = {'aps': {
            'alert': 'Alert!',
            'badge': 5,
            'sound': 'Meep Meep!',
            'content-available': 1,
            'category': 'stuff',
        }}
        self.assertEqual(frame.decoded, payload)

    def test_send_notification(self):
        msg = Message([_token1], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.assertEqual(self.sent_tokens, [_token1])

    def test_send_multiple(self):
        msg = Message([_token1, _token2], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.assertEqual(self.sent_tokens, [_token1, _token2])

    def test_reject_last(self):
        msg = Message([_token1, _token2], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.connection.set_inbuf(struct.pack('!BBI', 8, 1, self.sent_frames[-1].ident))

        sleep(0.1)

        self.assertEqual(self.connections_opened, 1)
        self.assertEqual(self.connections_closed, 1)
        self.assertEqual(self.sent_tokens, [_token1, _token2])
        self.assertEqual(self.apns_error.status, 1)

    def test_reject_middle(self):
        msg = Message([_token1, _token2, _token3], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.connection.set_inbuf(struct.pack('!BBI', 8, 1, self.sent_frames[-2].ident))

        sleep(0.1)

        self.assertEqual(self.connections_opened, 2)
        self.assertEqual(self.connections_closed, 1)
        self.assertEqual(self.sent_tokens, [_token1, _token2, _token3, _token3])
        self.assertEqual(self.apns_error.status, 1)

    def test_reject_unknown(self):
        msg = Message([_token1, _token2, _token3], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.connection.set_inbuf(struct.pack('!BBI', 8, 1, 100))

        sleep(0.1)

        self.assertEqual(self.connections_opened, 2)
        self.assertEqual(self.connections_closed, 1)
        self.assertEqual(self.sent_tokens, [_token1, _token2, _token3, _token1, _token2, _token3])
        self.assertEqual(self.apns_error, None)

    def test_shutdown(self):
        msg = Message([_token1, _token2, _token3], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.connection.set_inbuf(struct.pack('!BBI', 8, 10, self.sent_frames[-2].ident))

        sleep(0.1)

        self.assertEqual(self.connections_opened, 2)
        self.assertEqual(self.connections_closed, 1)
        self.assertEqual(self.sent_tokens, [_token1, _token2, _token3, _token3])
        self.assertEqual(self.apns_error, None)

    def test_read_exc(self):
        msg = Message([_token1], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.connection.set_inbuf(socket.error("Test error"))

        sleep(0.1)

        self.apns.send_message(msg)

        sleep(0.1)

        self.assertEqual(self.connections_opened, 2)
        self.assertEqual(self.connections_closed, 1)
        self.assertEqual(self.sent_tokens, [_token1, _token1])
        self.assertEqual(self.apns_error, None)

    def test_write_exc(self):
        self.apns
        self.connection.set_write_exc(socket.error("Test error"))
        msg = Message([_token1], {'aps': {'badge': 1}})
        self.apns.send_message(msg)

        sleep(0.1)

        self.assertEqual(self.connections_opened, 2)
        self.assertEqual(self.connections_closed, 1)
        self.assertEqual(self.sent_tokens, [_token1])
        self.assertEqual(self.apns_error, None)

    def test_flush(self):
        self.apns.send_aps([_token1], badge=1)
        self.apns.flush_messages()

        self.assertTrue(self.apns._queue.is_empty())

    def test_feedback_empty(self):
        self.connection_inbuf = b''
        self.apns.get_feedback(self.handle_feedback)

        sleep(0.1)

        self.assertEqual(len(self.feedbacks), 0)

    def test_feedback_partial(self):
        self.connection_inbuf = struct.pack('!IH30s', self._timestamp, 32, unhexlify(_token1))
        self.apns.get_feedback(self.handle_feedback)

        sleep(0.1)

        self.assertEqual(len(self.feedbacks), 0)

    def test_feedback_single(self):
        self.connection_inbuf = struct.pack('!IH32s', self._timestamp, 32, unhexlify(_token1))
        self.apns.get_feedback(self.handle_feedback)

        sleep(0.1)

        self.assertEqual(len(self.feedbacks), 1)
        self.assertEqual(self.feedbacks[0].token, _token1)
        self.assertEqual(self.feedbacks[0].when, self._when)

    def test_feedback_remainder(self):
        self.connection_inbuf = struct.pack('!IH32sI', self._timestamp, 32, unhexlify(_token1), 0)
        self.apns.get_feedback(self.handle_feedback)

        sleep(0.1)

        self.assertEqual(len(self.feedbacks), 1)
        self.assertEqual(self.feedbacks[0].token, _token1)
        self.assertEqual(self.feedbacks[0].when, self._when)

    def test_feedback_multiple(self):
        self.connection_inbuf = struct.pack(
            '!IH32sIH32sI',
            self._timestamp, 32, unhexlify(_token1),
            self._timestamp + 1, 32, unhexlify(_token2),
            0)
        self.apns.get_feedback(self.handle_feedback)

        sleep(0.1)

        self.assertEqual(len(self.feedbacks), 2)
        self.assertEqual(self.feedbacks[0].token, _token1)
        self.assertEqual(self.feedbacks[0].when, self._when)
        self.assertEqual(self.feedbacks[1].token, _token2)
        self.assertEqual(self.feedbacks[1].when, self._when + timedelta(seconds=1))

    #
    # Hooks
    #

    def new_connection(self, *args, **kwargs):
        connection = TestConnection(self, self.connection_inbuf)
        self.connections.append(connection)

        return connection

    def handle_error(self, error):
        self.apns_error = error

    def handle_feedback(self, feedback):
        self.feedbacks.append(feedback)

    #
    # State
    #

    @property
    def apns(self):
        if self._apns is None:
            self._apns = ApnsManager(
                'key-path', 'cert-path',
                backend_path='apns_worker.backend.threaded.Backend',
                message_grace=0.1, error_handler=self.handle_error
            )

        return self._apns

    @property
    def connection(self):
        return self.connections[-1] if len(self.connections) > 0 else None

    @property
    def connections_opened(self):
        return len([c for c in self.connections if c.is_opened])

    @property
    def connections_closed(self):
        return len([c for c in self.connections if c.is_closed])

    @property
    def sent_tokens(self):
        """ Tokens of send frames, in order. """
        return [frame.token for frame in self.sent_frames]

    @property
    def sent_idents(self):
        """ Idents of sent frames, in order. """
        return [frame.ident for frame in self.sent_frames]

    @property
    def sent_frames(self):
        """ All sent frames, in order. """
        return [frame for connection in self.connections for frame in connection.sent_frames]


class TestConnection(Connection):
    """ A fake apns_worker.backend.threaded.Connection. """
    def __init__(self, test_case, inbuf=None):
        self.test_case = test_case

        self.outbuf = six.BytesIO()
        self.inbuf = inbuf
        self.write_exc = None

        self._close_on_empty = (inbuf is not None)

        self.cond = Condition()
        self._is_opened = False
        self._is_closed = False

        self._sent_frames = None

    def __copy__(self):
        return self.test_case.new_connection()

    @property
    def is_opened(self):
        return self._is_opened

    def connect(self):
        self.is_opened = True

        return (not self._is_closed)

    def recv(self, bufsize=4096):
        """
        This always returns one byte at a time to make sure we're looping
        properly.
        """
        self._is_opened = True

        with self.cond:
            while self.inbuf is None:
                self.cond.wait()

            if isinstance(self.inbuf, Exception):
                val = self.inbuf
                self.inbuf = None
            else:
                val = self.inbuf[:1]
                self.inbuf = self.inbuf[1:] if (len(self.inbuf) > 1) else None
                if (self.inbuf is None) and self._close_on_empty:
                    self.close()

        if isinstance(val, Exception):
            raise val
        else:
            return val

    def sendall(self, buf):
        if self._is_closed:
            return 0

        self._is_opened = True
        self._sent_frames = None

        err = self.write_exc
        if err is not None:
            self.write_exc = None
            raise err
        else:
            return self.outbuf.write(buf)

    def close(self):
        with self.cond:
            self._is_closed = True
            self.inbuf = b''
            self.cond.notify_all()

    #
    # Test methods
    #

    def set_inbuf(self, val):
        """ Send data or an exception to the read thread. """
        with self.cond:
            self.inbuf = val
            self.cond.notify_all()

    def set_write_exc(self, exc):
        """ Set an exception for sendall. """
        self.write_exc = exc

    @property
    def sent_frames(self):
        """ Returns Frame objects representing sent frames. """
        if self._sent_frames is None:
            stream = six.BytesIO(self.outbuf.getvalue())
            self._sent_frames = list(takewhile(lambda f: f is not None, (Frame.parse(stream) for i in repeat(None))))

        return self._sent_frames


class Frame(namedtuple('Frame', ['token', 'payload', 'ident', 'expiration', 'priority'])):
    """ A parsed APNs frame. """
    @classmethod
    def parse(cls, stream):
        frame = None

        try:
            cmd, length = struct.unpack('!BI', stream.read(5))
            if cmd == 2:
                items = cls._parse_items(stream, length)
                frame = cls(**items)
        except struct.error:
            pass

        return frame

    @classmethod
    def _parse_items(cls, stream, length):
        items = {
            'token': None,
            'payload': None,
            'ident': None,
            'expiration': None,
            'priority': None,
        }

        while length > 0:
            item_id, data_len = struct.unpack('!BH', stream.read(3))
            data = stream.read(data_len)

            if item_id == 1:
                items['token'] = hexlify(data).decode('ascii')
            elif item_id == 2:
                items['payload'] = data
            elif item_id == 3:
                items['ident'] = struct.unpack('!I', data)[0]
            elif item_id == 4:
                items['expiration'] = struct.unpack('!I', data)[0]
            elif item_id == 5:
                items['priority'] = struct.unpack('!B', data)[0]
            else:
                raise ValueError("Unknown item ID in frame: {0}".format(item_id))

            length -= (3 + data_len)

        return items

    @property
    def decoded(self):
        return json.loads(self.payload.decode('utf-8'))



from setuptools import setup
import sys


setup(
    name='apns-worker',
    version='0.1',
    description="A robust client library for Apple's push notification service.",
    long_description=open('README.rst').read(),
    author='Peter Sagerson',
    author_email='psagers@getcloak.com',
    url='https://github.com/bbits/apns-worker',
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    install_requires=[
        'six',
    ],
    packages=[
        'apns_worker',
        'apns_worker.backend',
        'apns_worker.tests',
    ],
    include_package_data=True,
    zip_safe=False,

    test_suite='apns_worker.tests',
    tests_require=['mock'] if sys.version_info < (3, 3) else [],
)



"""
Datetime utilities to facilitate testing.
"""
from __future__ import unicode_literals, absolute_import


def now():
    """ Public API: import this. """
    return _now()


def _now():
    """ Private API: patch this. """
    import datetime

    return datetime.datetime.now()


def Now(dt):
    """ A context processor that sets the current datetime. """
    from datetime import date, time, datetime
    try:
        from unittest import mock
    except ImportError:
        import mock

    if isinstance(dt, date) and not isinstance(dt, datetime):
        dt = datetime.combine(dt, time(0))

    return mock.patch('{}._now'.format(__name__), lambda: dt)



from binascii import hexlify
from copy import copy
import logging
import os.path
import socket
import ssl
import struct
from threading import Thread, RLock, Condition
import time

import apns_worker
from apns_worker import apns

from . import base


logger = logging.getLogger(__name__)


class Backend(base.Backend):
    """
    A backend that uses the native Python concurrency model.

    This uses Python threads to interact with APNs.

    """
    def __init__(self, *args, **kwargs):
        super(Backend, self).__init__(*args, **kwargs)

        self.queue_cond = Condition()
        self.thread = None

    def start(self):
        self.thread = ReadThread(
            self, self.environment, self.key_path, self.cert_path,
            self.queue, self.queue_cond
        )
        self.thread.setDaemon(True)
        self.thread.start()

    def stop(self):
        if self.thread is not None:
            self.thread.terminate(wait=True)
            self.thread = None

    def start_feedback(self, callback):
        thread = FeedbackThread(self.environment, self.key_path, self.cert_path, callback)
        thread.start()

    def queue_lock(self):
        return self.queue_cond

    def queue_notify(self):
        self.queue_cond.notify_all()

    def sleep(self, seconds):
        time.sleep(seconds)


class ReadThread(Thread):
    """
    The master thread managing our APNs connection.

    This creates our connection to the service and waits for errors to arrive.
    It runs indefinitely, reconnecting as necessary.

    """
    def __init__(self, backend, environment, key_path, cert_path, queue, queue_cond):
        super(ReadThread, self).__init__()

        self.backend = backend
        self.connection = _new_connection(self._address(environment), key_path, cert_path)
        self.queue = queue
        self.queue_cond = queue_cond

        self.writer = None
        self._should_terminate = False

    def _address(self, environment):
        if environment == 'production':
            host = 'gateway.push.apple.com'
        else:
            host = 'gateway.sandbox.push.apple.com'

        return (host, 2195)

    def terminate(self, wait=True):
        with self.queue_cond:
            self._should_terminate = True
            self.connection.close()
            self.queue_cond.notify_all()

        if wait:
            self.join(1)
            if self.is_alive():
                logger.warning("Read thread did not terminate cleanly.")

    def run(self):
        logger.debug("Read thread starting.")

        while not self._should_terminate:
            try:
                self.wait_for_notification()
                if not self._should_terminate:
                    self.start_writing()
                    self.wait_for_error()
            except Exception as e:
                logger.warning("Uncaught exception in read thread: {0}".format(e))
            finally:
                self.reset()

        logger.debug("Read thread terminating.")

    def wait_for_notification(self):
        """ Blocks until we have a notification to send. """
        with self.queue_cond:
            while (not self.queue.has_unclaimed()) and (not self._should_terminate):
                self.queue_cond.wait()

    def start_writing(self):
        self.writer = WriteThread(self.connection, self.queue, self.queue_cond)
        self.writer.start()

    def wait_for_error(self):
        try:
            try:
                buf = self.connection.recv_min(6)
            finally:
                self.connection.close()
                self.stop_writing(wait=True)

            if len(buf) >= 6:
                self.handle_response_data(buf)
        except socket.error as e:
            logger.info("Socket error while reading: {0}.".format(e))
        except Exception as e:
            logger.warning("Exception while reading: {0}".format(e))

    def handle_response_data(self, buf):
        """ Process an error from the service. """
        try:
            _, status, ident = struct.unpack('!BBI', buf)
        except Exception as e:
            logger.warning("Failed to parse APNs response {0}: {1}".format(hexlify(buf), e))
        else:
            is_shutdown = (status == 10)
            notification = self.queue.backtrack(ident)
            if (notification is not None) and (not is_shutdown):
                error = apns.Error(status, notification.message, notification.token)
                logger.debug("Received response from push service: {0}".format(error))
                self.backend.delivery_error(error)

    def reset(self):
        self.connection.close()
        self.stop_writing(wait=False)
        self.connection = copy(self.connection)

    def stop_writing(self, wait=False):
        if self.writer is not None:
            self.writer.terminate(wait=wait)
            self.writer = None


class WriteThread(Thread):
    """
    A thread to pull notifications from the queue and send them over the wire.

    This is subordinate to the read thread, above. Once the connection is open,
    this thread is started and writes notifications indefinitely. When the
    socket closes, it terminates. The read thread will spawn a new write thread
    when it's ready to start writing again.

    """
    def __init__(self, connection, queue, queue_cond):
        super(WriteThread, self).__init__()

        self.connection = connection
        self.queue = queue
        self.queue_cond = queue_cond

        self._should_terminate = False

    def run(self):
        logger.debug("Write thread starting.")

        try:
            while not self._should_terminate:
                self.send_more_notifications()
        except socket.error as e:
            logger.info("Socket error while writing {0}.".format(e))
        except Exception as e:
            logger.warning("Exception while writing: {0}".format(e))
        finally:
            self.connection.close()

        logger.debug("Write thread terminating.")

    def send_more_notifications(self):
        notification = self.wait_for_notification()
        if notification is not None:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Sending {0}".format(notification))

            try:
                self.connection.sendall(notification.frame())
            except Exception:
                self.queue.unclaim(notification)
                raise

    def wait_for_notification(self):
        with self.queue_cond:
            notification = self.claim_notification()
            while (notification is None) and (not self._should_terminate):
                self.queue_cond.wait()
                notification = self.claim_notification()

        return notification

    def claim_notification(self):
        return (self.queue.claim() if (not self._should_terminate) else None)

    def terminate(self, wait=True):
        with self.queue_cond:
            self._should_terminate = True
            self.queue_cond.notify_all()

        if wait:
            self.join(1)
            if self.is_alive():
                logger.warning("Write thread did not terminate cleanly.")


class FeedbackThread(Thread):
    """
    Handles a connection to the feedback service.

    This makes a new connection to the feedback service and sends items to the
    callback as they arrive. It terminates when the connection is closed from
    the other end.

    """
    def __init__(self, environment, key_path, cert_path, callback):
        super(FeedbackThread, self).__init__()

        self.connection = _new_connection(self._address(environment), key_path, cert_path)
        self.callback = callback
        self.buf = b''

    def _address(self, environment):
        if environment == 'production':
            host = 'feedback.push.apple.com'
        else:
            host = 'feedback.sandbox.push.apple.com'

        return (host, 2196)

    def run(self):
        logger.debug("Feedback thread starting.")

        while not self.connection.is_closed:
            self.read_more()
            self.process_buffer()

        logger.debug("Feedback thread terminating.")

    def read_more(self):
        more = self.connection.recv()
        if len(more) > 0:
            self.buf += more
        else:
            self.connection.close()

    def process_buffer(self):
        feedback, remain = apns.Feedback.parse(self.buf)
        while feedback is not None:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Received feedback: {0}".format(feedback))

            self.callback(feedback)
            feedback, remain = apns.Feedback.parse(remain)

        self.buf = remain


def _new_connection(address, key_path, cert_path):
    """ Mock target. """
    return Connection(address, key_path, cert_path)


class Connection(object):
    """
    A thread-safe read-write connection to the APN service.

    The TCP connection will be opened the first time it's needed. Once the
    connection is closed, it can not be reopened. Use the copy module to create
    a shallow copy with a new uninitialized socket.

    """
    def __init__(self, address, key_path, cert_path):
        self.address = address
        self.key_path = key_path
        self.cert_path = cert_path

        self.lock = RLock()
        self._sock = None
        self._is_closed = False

    def __copy__(self):
        return self.__class__(self.address, self.key_path, self.cert_path)

    def connect(self):
        """ Force connection, if necssary. """
        return (self.sock() is not None)

    @property
    def is_closed(self):
        return self._is_closed

    def recv_min(self, count):
        """
        Reads a specific number of bytes.

        This won't return until the requested bytes are read or the connection
        is closed.

        :param int count: Minimum bytes to wait for.

        """
        buf = b''

        while len(buf) < count:
            more = self.recv(count - len(buf))
            if len(more) > 0:
                buf += more
            else:
                break

        return buf

    def recv(self, bufsize=4096):
        sock = self.sock()

        return sock.recv(bufsize) if (sock is not None) else b''

    def sendall(self, buf):
        sock = self.sock()

        return sock.sendall(buf) if (sock is not None) else 0

    def close(self):
        with self.lock:
            if self._sock is not None:
                try:
                    logger.debug("Closing connection to {0}.".format(self.address))
                    self._sock.close()
                finally:
                    self._sock = None

            self._is_closed = True

    def sock(self):
        with self.lock:
            if (self._sock is None) and (not self._is_closed):
                logger.debug("Opening connection to {0}.".format(self.address))
                sock = socket.create_connection(self.address)
                sock = ssl.wrap_socket(
                    sock, self.key_path, self.cert_path,
                    ssl_version=ssl.PROTOCOL_TLSv1, ca_certs=self.ca_certs()
                )
                self._sock = sock

        return self._sock

    def ca_certs(self):
        """ Path to the anchor certs. """
        return os.path.join(os.path.dirname(apns_worker.__file__), 'certs/anchors.pem')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from datetime import datetime
import unittest

from apns_worker.apns import Message


_token1 = '1ba97ad1311307c189696e2369c89fa83d652611a6e3c7370881289e45668fd3'
_token2 = '4c23f042050f48da350cb1079d61189cf7a47cb8df087429c0b2da65226cbecc'


class MessageTestCase(unittest.TestCase):
    def test_bad_token(self):
        with self.assertRaises(Exception):
            Message(['bogus'], {})

    def test_bad_payload(self):
        with self.assertRaises(Exception):
            Message([_token1], object())

    def test_bad_expiration(self):
        with self.assertRaises(Exception):
            Message([_token1], {}, expiration=5)

    def test_bad_priority(self):
        with self.assertRaises(Exception):
            Message([_token1], {}, priority='busted')

    def test_unicode_payload(self):
        Message([_token1], {'aps': {'alert': 'Ümlaut'}})

    def test_expiration(self):
        msg = Message([_token1], {}, expiration=datetime(2015, 1, 1))

        self.assertEqual(msg._encoded_expiration, 1420070400)

    def test_frame_1(self):
        msg = Message([_token1], {'aps': {'badge': 1}})
        frames = [notif.frame() for notif in msg.notifications()]

        self.assertEqual(
            frames,
            [b'\x02\x00\x00\x009\x01\x00 \x1b\xa9z\xd11\x13\x07\xc1\x89in#i\xc8\x9f\xa8=e&\x11\xa6\xe3\xc77\x08\x81(\x9eEf\x8f\xd3\x02\x00\x13{"aps":{"badge":1}}']
        )

    def test_frame_2(self):
        msg = Message([_token1], {'aps': {'badge': 1}}, expiration=datetime(2015, 1, 1), priority=5)
        frames = [notif.frame() for notif in msg.notifications()]

        self.assertEqual(
            frames,
            [b'\x02\x00\x00\x00D\x01\x00 \x1b\xa9z\xd11\x13\x07\xc1\x89in#i\xc8\x9f\xa8=e&\x11\xa6\xe3\xc77\x08\x81(\x9eEf\x8f\xd3\x02\x00\x13{"aps":{"badge":1}}\x04\x00\x04T\xa4\x8e\x00\x05\x00\x01\x05']
        )

    def test_frame_multiple(self):
        msg = Message([_token1, _token2], {'aps': {'badge': 1}}, expiration=datetime(2015, 1, 1), priority=5)
        frames = [notif.frame() for notif in msg.notifications()]

        self.assertEqual(
            frames,
            [
                b'\x02\x00\x00\x00D\x01\x00 \x1b\xa9z\xd11\x13\x07\xc1\x89in#i\xc8\x9f\xa8=e&\x11\xa6\xe3\xc77\x08\x81(\x9eEf\x8f\xd3\x02\x00\x13{"aps":{"badge":1}}\x04\x00\x04T\xa4\x8e\x00\x05\x00\x01\x05',
                b'\x02\x00\x00\x00D\x01\x00 L#\xf0B\x05\x0fH\xda5\x0c\xb1\x07\x9da\x18\x9c\xf7\xa4|\xb8\xdf\x08t)\xc0\xb2\xdae"l\xbe\xcc\x02\x00\x13{"aps":{"badge":1}}\x04\x00\x04T\xa4\x8e\x00\x05\x00\x01\x05',
            ]
        )



from abc import ABCMeta, abstractmethod

from six import add_metaclass


@add_metaclass(ABCMeta)
class Backend(object):
    """
    Base class for APNs backends.

    The arguments to __init__ are unspecified and subject to change. If a
    subclass wishes to add its own initialization, it must accept any arguments
    and pass them to the superclass.

    Backend instances will have key parameters set as instance attributes.

    .. attribute:: queue

        Our :class:`~apns_worker.queue.NotificationQueue`.

    .. attribute:: environment

        The APNs environment to talk to (`'sandbox'` or `'production'`).

    .. attribute:: key_path

        Path to our PEM-encoded TLS client key.

    .. attribute:: cert_path

        Path to our PEM-encoded TLS client certificate.

    """
    def __init__(self, queue, environment, key_path, cert_path, error_handler):
        self.queue = queue
        self.environment = environment
        self.key_path = key_path
        self.cert_path = cert_path
        self._error_handler = error_handler

        queue._set_backend(self)

    @abstractmethod
    def start(self):
        """
        Override this.

        Starts processing notifications.

        """

    @abstractmethod
    def stop(self):
        """
        Override this.

        Stops processing notifications.

        """

    @abstractmethod
    def start_feedback(self, callback):
        """
        Override this.

        Opens the APNs feedback connection. The callback will be called zero or
        more times with a :class:`~apns_worker.Feedback` object as the single
        argument.

        """

    @abstractmethod
    def queue_lock(self):
        """
        Override this.

        Returns an object compatible with :class:`threading.Lock`. Subclasses
        that don't require locking may return a dummy lock.

        """

    @abstractmethod
    def queue_notify(self):
        """
        Override this.

        Notifies listeners that the queue may have new items available.

        This is always called while the object returned by
        :meth:`~apns_worker.backend.base.Backend.queue_lock` is acquired,
        making it compatible with condition variable semantics.

        """

    @abstractmethod
    def sleep(self, seconds):
        """
        Override this.

        Sleeps for the given number of seconds.

        """

    def delivery_error(self, error):
        """
        Reports a permanent error delivering a message.

        :type message: :class:`~apns_worker.Message`
        :param str token: The specific token that failed.
        :type error: :class:`~apns_worker.Error`

        """
        if self._error_handler is not None:
            self._error_handler(error)
